#include "decode.h"
#include <stdio.h>
#include <chrono>
#include "lowpassfilter.h"
#include <assert.h>
#include <fstream>
#include "helpers.h"
#include "matched_filter.h"

#define PI 3.14159265


void decode(string signal_file_name, string output_file_name, unsigned long sample_rate, long carrier_frequency, unsigned long samples_per_bit, long periods_per_symbol, long overlapping_samples){
    
    std::ofstream decoded_message (output_file_name);
    
    //word rate is 1/4 the bit rate because we send 4 bits per word
    double symbol_width = samples_per_bit * 4;

    //read all the input data. Possibly preallocate if we can set parameters in the text file
    double c;
    std::vector<double> rcvd_signal;
    
    std::ifstream myfile (signal_file_name);
    while(myfile >> c){
        rcvd_signal.push_back(c);
    }
    myfile.close();
    
    //recover raw (in phase and quadrature) components
    //Generate Q and I signals
    std::vector<double> raw_quadrature_stream;
    std::vector<double> raw_in_phase_stream;
    
    raw_quadrature_stream.resize(rcvd_signal.size());
    raw_in_phase_stream.resize(rcvd_signal.size());
    double freq_ratio = (double) carrier_frequency / (double) sample_rate;
    
    for (long i = 0; i < rcvd_signal.size(); ++i){
        raw_in_phase_stream[i] = rcvd_signal[i] * sin(2 * PI * freq_ratio * i);
        raw_quadrature_stream[i] = rcvd_signal[i] * sin(2 * PI * freq_ratio * i + PI / 2);
    }
    
    //low pass filter
    //double frequency terms are at f = 0.4 * PI, information is at 0.2 * PI
    //hence let the cutoff frequency be 0.3 * PI
    
    //filter out double frequency terms
    
    std::vector<double> filtered_quadrature_stream;
    std::vector<double> filtered_in_phase_stream;
    
    //make vectors appropriate size
    filtered_quadrature_stream.resize(raw_quadrature_stream.size() + low_pass_filter_length - 1);
    filtered_in_phase_stream.resize(raw_in_phase_stream.size() + low_pass_filter_length - 1);
    
    //I used MATLAB to generate a simple FIR filter with group delay of 10 samples
    convolve(raw_quadrature_stream, lpf_vector, filtered_quadrature_stream);
    convolve(raw_in_phase_stream, lpf_vector, filtered_in_phase_stream);
    
    //double the magnitude of the signal to account for modulation and demodulation
    for (long i = 0; i < filtered_quadrature_stream.size(); ++i){
        filtered_quadrature_stream[i] = filtered_quadrature_stream[i] * 2;
        filtered_in_phase_stream[i] = filtered_in_phase_stream[i] * 2;
    }
 
    
    filtered_in_phase_stream.erase(filtered_in_phase_stream.begin(), filtered_in_phase_stream.begin() + low_pass_filter_length - 1);
    filtered_quadrature_stream.erase(filtered_quadrature_stream.begin(), filtered_quadrature_stream.begin() + low_pass_filter_length - 1);

    vector<vector<double> > stacked_signals;

    //stack_interleaving_signals pairs each Q and I signal together for the sake of decoding
    //In: [Q1 Q2 Q3 ... QN]
    //    [I1 I2 I3 ... IN]
    //Out: [[I1 Q1] [I2 Q2] [I3 Q3] ... [IN QN]]
    stack_interleaving_signals(filtered_in_phase_stream, filtered_quadrature_stream, symbol_width, periods_per_symbol, overlapping_samples, stacked_signals);

    bool is_srrc = 1; //using a square root raised cosine filter
    Matched_Filter_Bank filter_bank(symbol_width, periods_per_symbol, is_srrc);
    vector<uint8_t> codewords_QAM;
    codewords_QAM.resize(stacked_signals.size());
    
    vector<uint8_t> codewords_bits;
    codewords_bits.resize(codewords_QAM.size() / 2);
    
    //convolve with the matched filter
    chrono::time_point<std::chrono::system_clock> matched_filter_start_time, matched_filter_end_time; // timing for matched filter operation
    matched_filter_start_time = chrono::system_clock::now();
    matched_filter_handler(stacked_signals, filter_bank, codewords_QAM, symbol_width * periods_per_symbol - overlapping_samples, sample_rate);
    matched_filter_end_time = chrono::system_clock::now();
    cout << "Time for matched filter operations : " << chrono::duration<double>(matched_filter_end_time - matched_filter_start_time).count() << "s\n";


    gray_to_bits(codewords_QAM, codewords_bits);

    for (int i = 0; i < codewords_bits.size(); ++i) {
        decoded_message << (char) codewords_bits[i];
    }
    
    cout << endl;
        
}
