#include "generate_signal.h"
#include <chrono>
#include <assert.h>
#include <math.h>
#include <vector>
#include <fstream>
#include "helpers.h"
#include "matched_filter.h"
#include "AWGN.h"

using namespace std;


void generate_signal(string input_file_name, string output_file_name, unsigned long sample_rate, long carrier_frequency, long samples_per_bit, long periods_per_symbol, long overlapping_samples, double SNR) {
    
    ifstream message_file (input_file_name);
    ofstream myfile (output_file_name);

    //read the message file
    char c;
    
    //vector to hold the bits
    vector<bool> bits;
    
    //order it from LSB to MSB
    while (message_file.get(c)){
        bits.push_back((c & 0x80) >> 7);
        bits.push_back((c & 0x40) >> 6);
        bits.push_back((c & 0x20) >> 5);
        bits.push_back((c & 0x10) >> 4);
        bits.push_back((c & 0x8) >> 3);
        bits.push_back((c & 0x4) >> 2);
        bits.push_back((c & 0x2) >> 1);
        bits.push_back(c & 0x1);
    }
    

    //figure out length of message
    unsigned long number_of_bits = bits.size();
    //4 bits = 1 codeword
    unsigned long number_of_cw = number_of_bits / 4;
    
    //Generate Q and I constellation points
    vector<double> in_phase_vals;
    vector<double> quadrature_vals;
    
    in_phase_vals.resize(number_of_bits / 4);
    quadrature_vals.resize(number_of_bits / 4);
    
    for(int i = 0; i < number_of_bits; i += 4){
        pair<double, double> point = bits_to_QAM(bits[i], bits[i + 1], bits[i + 2], bits[i + 3]);
        in_phase_vals[i / 4] = point.first;
        quadrature_vals[i / 4] = point.second;
    }
    

    //build "analog" signal
    //note, the length of each pulse is the pulse_width + overlapping_samples
    long symbol_width = samples_per_bit * 4; //period of the pulse
    bool is_srrc = 1;
    Matched_Filter_Bank signal_bank(symbol_width, periods_per_symbol, is_srrc);

    long num_samples = symbol_width * periods_per_symbol * number_of_cw + (periods_per_symbol - 1) * symbol_width * periods_per_symbol; // divide by four because 4 bits = 1CW
    
    //Generate Q and I signals from constellation points
    vector<double> in_phase_signal(num_samples, 0);
    vector<double> quadrature_signal(num_samples, 0);
    
    //where the Q symbol stops and the I symbol begins
    long QI_division = symbol_width * periods_per_symbol;
    //long begin_overlap = symbol_width*(periods_per_symbol - 1)/2 - overlapping_samples;
    
    //first codeword is different from the others because it has no overlap on the front
    int symbol = (bits[0] << 3) | (bits[1] << 2) | (bits[2] << 1) | (bits[3]);
    vector<double> mysignal = signal_bank.filters[symbol];
    for (int j = 0; j < symbol_width * periods_per_symbol; ++j){
        in_phase_signal[j] = signal_bank.filters[symbol][j];
        quadrature_signal[j] = signal_bank.filters[symbol][j + QI_division];
    }
    //all subsequent signals have overlap on the front and back now
    for (int i = 4; i < number_of_bits; i +=4) {
        symbol = (bits[i] << 3) | (bits[i+1] << 2) | (bits[i+2] << 1) | (bits[i+3]);
        mysignal = signal_bank.filters[symbol];
        for (int j = 0; j < symbol_width * periods_per_symbol; ++j){
            in_phase_signal[i/4 * (symbol_width * periods_per_symbol - overlapping_samples) + j] += signal_bank.filters[symbol][j];
            quadrature_signal[i/4 * (symbol_width * periods_per_symbol - overlapping_samples) + j] += signal_bank.filters[symbol][j + QI_division];
        }
    }

    //signal vector
    vector<double> noisy_signal;
    noisy_signal.resize(num_samples); // divide by two for two channels (Q and I)
    //store result here for easier calculation
    double freq_ratio = (double) carrier_frequency / (double) sample_rate;
    
    
    //generate noise. Avg Signal Power of 16 QAM is 10-> SNR = 10 * log10(10/N)-> N = 10^(1-SNR)
    double noise_power = pow(10, 1-SNR/10);
    long samples = noisy_signal.size();
    
    //Generate Random Input Signal of 0's and 1's
    long long seed = chrono::system_clock::now().time_since_epoch().count();
    default_random_engine generator ((unsigned int) seed);
    uniform_int_distribution<int> distribution(0,1);
    
    AWGN g(0.0,noise_power,samples);
    double * noise = g.generateNoiseSamples();
    
    //modulate the signal
    for(int i = 0; i < noisy_signal.size(); ++i){
        noisy_signal[i] = sin(2 * PI * freq_ratio * i) * in_phase_signal[i]
        + sin(2 * PI * freq_ratio * i + PI / 2) * quadrature_signal[i] + noise[i];
        myfile << noisy_signal[i] << " ";
    }

    g.deallocate();
}
