//
//  helpers.cpp
//  MatchedFilter
//
//  Created by Nathan Block on 1/24/20.
//  Copyright © 2020 Nathan Block. All rights reserved.
//

#include "helpers.h"
using namespace std;



/*converts the two bits input to a QAM signal
 0b11 -> 3
 0b10 -> 1
 0b01 -> -1
 0b00 -> -3
 
 */

//Converts the four bits defined in the input into a QAM signal with a Gray Code
pair<int, int> bits_to_QAM(int first, int second, int third, int fourth){
    int code_word = fourth | third << 1 | second << 2 | first << 3;
    pair <double,double> point; //first is in-phase, second is quadrature
    switch (code_word){
        case 0:
            point.first = -3;
            point.second = -3;
            break;
        case 1:
            point.first = -3;
            point.second = -1;
            break;
        case 2:
            point.first = -3;
            point.second = 3;
            break;
        case 3:
            point.first = -3;
            point.second = 1;
            break;
        case 4:
            point.first = -1;
            point.second = -3;
            break;
        case 5:
            point.first = -1;
            point.second = -1;
            break;
        case 6:
            point.first = -1 ;
            point.second = 3 ;
            break;
        case 7:
            point.first = -1;
            point.second = 1;
            break;
        case 8:
            point.first = 3;
            point.second = -3;
            break;
        case 9:
            point.first = 3;
            point.second = -1;
            break;
        case 10:
            point.first = 3;
            point.second = 3;
            break;
        case 11:
            point.first = 3;
            point.second = 1;
            break;
        case 12:
            point.first = 1;
            point.second = -3;
            break;
        case 13:
            point.first = 1;
            point.second = -1;
            break;
        case 14:
            point.first = 1;
            point.second = 3;
            break;
        case 15:
            point.first = 1;
            point.second = 1;
            break;
    }
    return point;
}

//turns 16 QAM gray coded signal into bits
void gray_to_bits(const vector<uint8_t> gray, vector<uint8_t> & data){
    
    for (int i = 0; i < gray.size(); i += 2) {
        data[i/2] = (gray[i] << 4) + gray[i + 1];
    }
    
}

//when not using the matched filter, this converts to bits using the nearest constellation, only designed for 16 QAM optimal boundaries though
void QAM_to_bits(const vector<double> & in_phase, const vector<double> & quad, vector<uint8_t> & data){

    vector<uint8_t> bits;
    bits.resize(in_phase.size());
    
    for(int i = 0; i < in_phase.size(); ++i){
        if((quad[i] > 2) && (in_phase[i] > 2)){
            bits[i] = 10;
        }
        else if((quad[i] > 0) && (in_phase[i] > 2)){
            bits[i] = 14;
        }
        else if((quad[i] > -2) && (in_phase[i] > 2)){
            bits[i] = 6;
        }
        else if((quad[i] < -2) && (in_phase[i] > 2)){
            bits[i] = 2;
        }
        else if((quad[i] > 2) && (in_phase[i] > 0)){
            bits[i] = 11;
        }
        else if((quad[i] > 0) && (in_phase[i] > 0)){
            bits[i] = 15;
        }
        else if((quad[i] > -2) && (in_phase[i] > 0)){
            bits[i] = 7;
        }
        else if((quad[i] < -2) && (in_phase[i] > 0)){
            bits[i] = 3;
        }
        else if((quad[i] > 2) && (in_phase[i] > -2)){
            bits[i] = 9;
        }
        else if((quad[i] > 0) && (in_phase[i] > -2)){
            bits[i] = 13;
        }
        else if((quad[i] > -2) && (in_phase[i] > -2)){
            bits[i] = 5;
        }
        else if((quad[i] < -2) && (in_phase[i] > -2)){
            bits[i] = 1;
        }
        else if((quad[i] > 2) && (in_phase[i] < -2)){
            bits[i] = 8;
        }
        else if((quad[i] > 0) && (in_phase[i] < -2)){
            bits[i] = 12;
        }
        else if((quad[i] > -2) && (in_phase[i] < -2)){
            bits[i] = 4;
        }
        else if((quad[i] < -2) && (in_phase[i] < -2)){
            bits[i] = 0;
        }
    }
    //convert sequences of 4 bits into 8 bits for readability
    for (int i = 0; i < bits.size(); i += 2) {
        data[i/2] = (bits[i] << 4) + bits[i + 1];
    }
}

void compute_error_rates(string sent_message_filename, string rcvd_message_filename, double SNR_dB){
    ifstream sent (sent_message_filename);
    ifstream rcvd (rcvd_message_filename);
    
    vector<char> sent_chars;
    vector<char> rcvd_chars;
    
    char c;
    while (sent.get(c)){
        sent_chars.push_back((uint8_t) c);
    }
    while (rcvd.get(c)){
        rcvd_chars.push_back((uint8_t) c);
    }
    
    int byte_error_count = 0;
    unsigned long byte_count = min(sent_chars.size(), rcvd_chars.size());
    int symbol_error_count = 0;
    unsigned long symbol_count = byte_count * 2;
    unsigned long bit_count = byte_count * 8;;
    int bit_error_count = 0;
    
    //compute bit error rate
    for (int i = 0; i < byte_count; ++i) {
        for (int mask = 0; mask < 8; ++mask){
            if((((rcvd_chars[i] & (1 << mask)) >> mask) & 1) != (((sent_chars[i] & (1 << mask)) >> mask) & 1)){
                ++bit_error_count;
            }
        }
    }
    
    //compute bit error rate
    cout << "Bits Received : " << bit_count <<endl;
    cout << "Bit Errors : " << bit_error_count<<endl;
    cout << "Bit Error Rate : " << ((double) bit_error_count) / ((double) bit_count)<<endl;
    
    //calculate symbol error rate
    for (int i = 0; i < byte_count; ++i) {
        if((sent_chars[i] & 0x0F) != (rcvd_chars[i] & 0x0F)){
            ++symbol_error_count;
        }
        if((sent_chars[i] & 0xF0) != (rcvd_chars[i] & 0xF0)){
            ++symbol_error_count;
        }
    }
    
    //compute byte error rate
    cout << "Symbols Received : " << symbol_count <<endl;
    cout << "Symbol Errors : " << symbol_error_count<<endl;
    cout << "Symbol Error Rate : " << ((double) symbol_error_count) / ((double) symbol_count)<<endl;
    
    /*
    double SNR_linear = pow(10, SNR_dB/10);
    double N0 = 2 * 10/SNR_linear; //signal power is 10, SNR = Signal Power / Noise Power, N0 = noise power * 2
    double Eb = (double) 10 / 4; //average energy per symbol is 10, 4 symbols per bit
    double second_try = (3.0/4.0) * (0.5) * erfc(sqrt(4.0/5.0*Eb/N0) / sqrt(2.0)) - pow((3.0/4.0) * (0.5) * erfc(sqrt(4.0/5.0*Eb/N0) / sqrt(2.0)),2);
     
    cout << "Theoretical Symbol Error Rate Bound : e <= " << second_try << endl;
    */
    //calculate byte error rate
    for (int i = 0; i < byte_count; ++i) {
        if(sent_chars[i] != rcvd_chars[i]){
            ++byte_error_count;
        }
    }
    
    //compute byte error rate
    cout << "Bytes Received : " << byte_count <<endl;
    cout << "Byte Errors : " << byte_error_count<<endl;
    cout << "Byte Error Rate : " << ((double) byte_error_count) / ((double) byte_count)<<endl;
    
    
}
