//
//  matched_filter.cpp
//  MatchedFilter
//
//  Created by Nathan Block on 1/27/20.
//  Copyright © 2020 Nathan Block. All rights reserved.
//

#include "matched_filter.h"
#include <iostream>

using namespace std;

#define PI 3.141519

//default constructor is for 16QAM
Matched_Filter_Bank::Matched_Filter_Bank(long symbol_width, long periods_per_symbol, bool is_srrc){
    num_filters = 16;
    filters.resize(num_filters);
    squared_magnitudes.resize(num_filters);
    vector<vector<double> > signal_set;
    signal_set.resize(4);
    

    
    vector<double> plus_ones(symbol_width * periods_per_symbol, 1);
    vector<double> plus_threes(symbol_width * periods_per_symbol, 3);
    vector<double> minus_ones(symbol_width * periods_per_symbol, -1);
    vector<double> minus_threes(symbol_width * periods_per_symbol, -3);
    vector<double>srrc(symbol_width * periods_per_symbol, 0);

    if(is_srrc){
        double a = 0.9;
        double T = symbol_width;
        double P = 10/T;

        long start = -(symbol_width * periods_per_symbol) / 2;
        long end = (symbol_width * periods_per_symbol) / 2;
        for (long i = start; i < end; ++i) {
            if(i == 0){
                srrc[i -start] = srrc[i - 1 - start]; // take care of the divide by zero case
            }
            else if(i == start){ //take care of t = 0 case
                srrc[i - start] = P * (4*a)/PI * sqrt(T) *
                (cos(PI * (1+a)*(i+1)/T) + sin(PI * (1-a)*(i+1)/T)/(4*a*(i+1)/T))/(1-pow((4*a*(i+1)/T),2));
            }
            else{
                srrc[i - start] = P * (4*a)/PI * sqrt(T) *
                (cos(PI * (1+a)*i/T) + sin(PI * (1-a)*i/T)/(4*a*i/T))/(1-pow((4*a*i/T),2));
            }
        }
    }

        
    for (long i = 0; i < symbol_width * periods_per_symbol; ++i) {
        plus_ones[i] *= srrc[i];
        minus_ones[i] *= srrc[i];
        plus_threes[i] *= srrc[i];
        minus_threes[i] *= srrc[i];
    }
    
    signal_set[0] = minus_threes;
    signal_set[1] = minus_ones;
    signal_set[2] = plus_threes;
    signal_set[3] = plus_ones;
    
    for (int i = 0; i < 4; ++i) {
        for (int j = 0; j < 4; ++j) {
            filters[4 * i + j] = signal_set[i];
            filters[4 * i + j].insert(filters[4 * i + j].end(), signal_set[j].begin(), signal_set[j].end());
        }
    }
    
    for (int i = 0; i < 16; ++i) {
        squared_magnitudes[i] = magnitude_squared(filters[i]);
    }
    
    double avg_signal_power = sqrt(10 * magnitude_squared(srrc) / symbol_width);
    
    //this calculation is needed for the SNR's to work out
    //the average power of 16QAM is sqrt(10), multiply this by the magnitude of each srrc pulse to
    //obtain the power per codeword. This needs to be multiplied by the length of each codeword to account
    //for the fact that noise power is calculated per sample
    for (int i = 0; i < 16; ++i) {
        for (int j = 0; j < filters[i].size(); ++j) {
            filters[i][j] /= avg_signal_power;
        }
    }
    //we've normalized now
    for (int i = 0; i < 16; ++i) {
        squared_magnitudes[i] = magnitude_squared(filters[i]);
    }
}

double magnitude_squared(const vector<double> & signal){
    long N = signal.size();
    double mag = 0;
    for (int i = 0; i < N; ++i) {
        mag += pow(signal[i],2);
    }
    return mag;
}

void convolve(const vector<double> & signal1, const vector<double> & signal2, vector<double> & output){
    long N1 = signal1.size();
    long N2 = signal2.size();
    for (long i = 0; i < N1 + N2 - 1; i++){
        output[i] = 0;
        for (long j = 0; j < N2; j++){
            if((i - j >= 0) && (i - j < N1)){ //make sure we're not indexing out of bounds
                output[i] += signal1[i - j] * signal2[j];
            }
        }
    }
}

double correlate(const vector<double> & first, const vector<double> & second){
    //requires that first and second are the same size
    //note that we don't want an assert in a timed code portion, so we _assume_ this is true. IF YOU ARE WRITING A TESTCASE, DO NOT BREAK THAT ASSUMPTION!
    //assert(first.size() == second.size());
    double out = 0;
    for (int i = 0; i < first.size(); ++i) {
        out += first[i] * second[i];
    }
    return out;
    
    
}


// interleave the signals so we get a vector containing [in1 quad1 in2 quad2 in3 quad3 ...inN quadN]
void stack_interleaving_signals(const vector<double> & raw_in_phase_stream, const vector<double> & raw_quadrature_stream, double symbol_width, long periods_per_symbol, long overlapping_samples, vector<vector<double> > & stacked_signals){
    
    int offset = 0;
    int symbol_length = (symbol_width * periods_per_symbol);
    
    int code_word_index = 0;
    while(offset < raw_in_phase_stream.size() - symbol_length) {
        vector<double> placeholder;
        stacked_signals.push_back(placeholder);
        stacked_signals[code_word_index].insert(stacked_signals[code_word_index].begin(), raw_quadrature_stream.begin() + offset, raw_quadrature_stream.begin() + symbol_length + offset);
        stacked_signals[code_word_index].insert(stacked_signals[code_word_index].begin(), raw_in_phase_stream.begin() + offset, raw_in_phase_stream.begin() + symbol_length + offset);
        offset += (symbol_width * periods_per_symbol - overlapping_samples);
        ++code_word_index;
    }
}





void matched_filter_handler(const vector<vector<double> > & stacked_signal, const Matched_Filter_Bank & my_filters, vector<uint8_t> & code_words, long cw_rate, long sample_rate){
    
    long codeword_index = 0;
    for(long sample_index = 0; sample_index < stacked_signal.size(); sample_index += 1) {
        code_words[sample_index] = run_matched_filter(stacked_signal[codeword_index], my_filters, sample_index, cw_rate);
        ++codeword_index;
    }

    
}

double mag_diff(const vector<double> & received_signal, const vector<double> & expected_signal){
    double total = 0;
    unsigned long N = received_signal.size();
    for (int i = 0; i<N; ++i) {
        total += abs(received_signal[i] - expected_signal[i]);
    }
    return total;
}

int run_matched_filter(const vector<double> & stacked_signal, const Matched_Filter_Bank & my_filters, long sample_index, long cw_rate){
    //make vectors appropriate size
    
    int best_guess = 0;
    double this_guess_val = 0;
    double best_guess_val = mag_diff(stacked_signal, my_filters.filters[0]);
    for(int i = 1; i < my_filters.num_filters; ++i){
        this_guess_val = mag_diff(stacked_signal, my_filters.filters[i]);
        if(this_guess_val < best_guess_val){
            best_guess_val = this_guess_val;
            best_guess = i;
        }
    }
    return best_guess;
}
